! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: edet2sky_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task transforms sky coordinates to either DETX/Y or RAWX/Y.
! 
! Comments:
!	TODO:
!	- Write units and comment to any new columns.
!	- Overwrite old INSTRUME kwd??
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module edet2sky_mod

  use dal
  use param  
  use errorHandling
  use cal		! -> BoresightT
  use caloaldefs

!  use caloalutils ! ???? doesn't this have the instId functions? 
!  But edet2sky seems to work without it! Maybe there is some link via the cal.

  use cal_aux	! for getCalInfo
  use constants
  use coordinate	! for raDecToInst and instToRaw;
			! also for detUnitArcsec definition.
  use dal_aux	! for readColDataToFixed, splitSetTabName, addOrOpenColumn.
!use geometric_types !*** temporary
  use ssc_misc	! for intToStr
  use wcs_aux           ! for type(WcsT)

  implicit none 

  ! Command-line parameters:

  logical ::&
    withHeader = .false.,&
    withOutSet = .false.
    !checkForFov = .true.
  character(255) ::&
    inSetTabName = '',&
    outSetTabName = '',&
    calInfoSetName = ''
  character(20) ::&
    dataStyle = '',&
    inputUnit = '',&
    userInstrument = '',&
    mosNodeStr = '',&
    calInfoStyle = '',&
    userDateTimeStr = ''
  real(single) ::&
    userCamX = 0.0,&
    userCamY = 0.0,&
    userScAttitudeRaDeg = 0.0,&
    userScAttitudeDecDeg = 0.0,&
    userScAttitudeAposDeg = 0.0

  character(255) ::&
    inSetName = '',&
    outSetName = '',&
    inTabName = '',&
    outTabName = ''
  character(20), parameter :: myDefaultStr = 'null_default'

  integer :: userCcd = 1,& 
             OFF_CHIP_BIT = 0

  private 

  public edet2sky

contains 

  !---------------------------------------------------------------------
  subroutine edet2sky
  !---------------------------------------------------------------------

    implicit none

    real(double), dimension(:), pointer :: camX, camY
    integer(int8), dimension(:), pointer :: ccdIn, ccdNum, ccdNumdum

    real(double),   dimension(:), allocatable :: detX, detY
    real(double),   dimension(:), allocatable :: ra, dec
    real(double), dimension(:), allocatable :: rawX, rawY
    real(double), dimension(:), allocatable :: rawXdum, rawYdum
    integer(int32) :: vsize
    integer(int32) :: instrumentId
    real(double) :: localTimeStamp
    type(SpacecraftAttitudeType) :: scAttitude, instrumAttitude
    logical(bool), allocatable :: offChip(:)
    integer :: mosNode
    type(WcsT)   :: wcs_image, wcs_sky ! defined in ssclib/wcs_aux
    real(single), dimension(:), allocatable :: xImagePix, yImagePix,&
                                               xSkyPix, ySkyPix


    !-------------------------------------------------------------------
    ! Read the input parameters 
    call readParameters

    !...................................................................
    ! Get the input data:
    call readCamera(camX, camY, ccdIn)

    !...................................................................
    ! Cal setups:
    call message(AppMsg, Verbose, 'CAL is being set up.')

    !   Turn off randomisation - important for conversion from RAW coords
    call CAL_setState(randomize=.false.)

    select case(calInfoStyle)
    case('set')
      call getCalInfo(calInfoSetName, instrumentId, localTimeStamp&
        , scAttitude, instrumAttitude) ! in ssclib/cal_aux. This also
      ! sets the instrument state of the cal.

      call getWcsKeys(calInfoSetName, wcs_image, wcs_sky)

    case('user')
      instrumentId = instId(userInstrument)
      localTimeStamp = toTimeTag(userDateTimeStr)
      ! in caloalutils/caloaldefs.f90

      scAttitude%ra   = userScAttitudeRaDeg   * PiOn180
      scAttitude%dec  = userScAttitudeDecDeg  * PiOn180
      scAttitude%apos = userScAttitudeAposDeg * PiOn180

      call CAL_setState(instrument=instrumentId)

    case default
      call error('badCalInfoStyle', "The value of the --calinfostyle &
                 &parameter was not recognised.")
    end select

    ! set the MOS node if necessary
    if (instrumentId .eq. EMOS1 .or. instrumentId .eq. EMOS2) then
      if (mosNodeStr .eq. 'primary') then
        mosNode = EMOS_PRIMARY_NODE
      else if (mosNodeStr .eq. 'redundant') then
        mosNode = EMOS_REDUNDANT_NODE
      else
        call error('badNodeStr', 'The mos node string '//trim(mosNodeStr)&
                 //' was not recognised.')
      end if
      call CAL_setState(ccdNodeId=mosNode)
    end if

!    write(errstr, '(a, f12.2, a)') 'Timestamp = ', myTimeStamp, ' sec.'
!    call message(AppMsg, Verbose)

    !...................................................................
    ! Now do the actual transformation (then output):

    ! Set up the output arrays
    vsize = size(camX)
    allocate(ra(vsize), dec(vsize))
    allocate(rawX(vsize), rawY(vsize), ccdNum(vsize), offChip(vsize))
    allocate(detX(vsize), detY(vsize))
    allocate(xImagePix(vsize), yImagePix(vsize))
    allocate(xSkyPix(vsize), ySkyPix(vsize))

    call message(AppMsg, Verbose, 'Transforming coordinates.')

    ! Fill the RAW and DET coord arrays
    select case(inputUnit)
    case('raw')

      rawX=camX 
      rawY=camY
      ccdNum=ccdIn
      call eSkyrawToDet(rawX, rawY, ccdNum, detX, detY) 
      ! Find out if input position lies on a CCD
      allocate(rawXdum(vsize), rawYdum(vsize), ccdNumdum(vsize))
      call detToRaw(detX, detY, rawXdum, rawYdum, offChip, ccdNumdum)

    case('det')

      detX=camX
      detY=camY
    
      call message(AppMsg, Verbose, 'Transforming into RAW.')
      call detToRaw(detX, detY, rawX, rawY, offChip, ccdNum)

    case default
      call error('badInpUnit', "The value of the --inputunit parameter &
                 &was not recognised.")
    end select

    ! Transform detector to celestial coords
    call detToRaDec(detX, detY, scAttitude, localTimeStamp,&
                      ra, dec)

    ! If WCS obtained from file, transform to image and sky coords
    if (calInfoStyle .eq. 'set') then

      ! Transform to image pixel coords
      call message(AppMsg, Verbose, 'Transforming into image pixels.')

      ! If an X/Y image convert from ra/dec, otherwise convert from det coords
      select case(wcs_image%x%ctype)
      case('RA---TAN')
       call raDecToPixels(ra, dec, wcs_image, xImagePix, yImagePix)
      case('DETX')
       call linearToImPixels(detx, dety, wcs_image, xImagePix, yImagePix)
      case default
       call warning('NoImagePix',"Can not calculate the image pixel")
      end select

      ! Transform celestial to sky (X/Y) pixels
      call message(AppMsg, Verbose, 'Transforming into sky pixels.')
      call raDecToPixels(ra, dec, wcs_sky, xSkyPix, ySkyPix)

    end if

    ! Finally output the results
    call makeOutput(rawX, rawY, offChip, ccdNum, detX, detY,&
                    ra, dec, xImagePix, yImagePix,&
                    xSkyPix, ySkyPix, instrumentId)

    ! Close down and lets go home
    deallocate(rawX, rawY, offChip, ccdNum)
    deallocate(detX, detY)
    deallocate(ra, dec)

  end subroutine edet2sky

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading parameters.')

    ! Find the units of the input - from user or set
    inputUnit = stringParameter('inputunit')

    dataStyle = stringParameter('datastyle')
    select case(dataStyle)
    case('set')
      inSetTabName = stringParameter('intab')
      !raColName          = 'RA'
      !decColName         = 'DEC'
      withOutSet = booleanParameter('withouttab')
      outSetTabName = stringParameter('outtab')
      !rawXcolName        = 'RAWX'
      !rawYcolName        = 'RAWY'
      !ccdNrColName       = 'CCDNR'
      !detXcolName        = 'DETX'
      !detYcolName        = 'DETY'
!***** move these initialisations to the header.

    case('user')
      userCamX = realParameter('X')
      userCamY = realParameter('Y')

      withHeader = booleanParameter('withheader')

      select case(inputUnit)
      case('det')
      case('raw')
        userCcd = intParameter('ccd')
      case default
        call error('badInpUnit',"The value of the --inputunit parameter &
                   &was not recognised.")
      end select

    case default
        call error('badDataStyle',"The value of the --datastyle parameter &
                   &was not recognised.")
    end select

    calInfoStyle = stringParameter('calinfostyle')
    select case(calInfoStyle)
    case('set')
      calInfoSetName = stringParameter('calinfoset')

    case('user')
      userInstrument = stringParameter('instrument')
      userDateTimeStr = stringParameter('datetime')
      userScAttitudeRaDeg = realParameter('scattra')
      userScAttitudeDecDeg = realParameter('scattdec')
      userScAttitudeAposDeg = realParameter('scattapos')

    case default
     call error('badCalInfoStyle',"The value of the --calinfostyle &
                &parameter was not recognised.")
    end select

    mosNodeStr = stringParameter('mosccdnode')

    applyAttcalcFudge = booleanParameter('withboresightfudge')
    ! only needed while boresight psi has wrong sign. IMS 2005-11-17

    !checkForFov  = booleanParameter('checkfov')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine readCamera(camX, camY, ccdNum)
  !---------------------------------------------------------------------

  ! This routine reads camera coordinates - either DETX/Y or RAWX/Y+CCD
    implicit none

    real(double), dimension(:), pointer :: camX, camY
    integer(int8), dimension(:), pointer :: ccdNum

    integer(int32) :: vsize
    !integer(int16) :: vsize
    type(DataSetT) :: set
    type(TableT) :: tab

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Reading input data.')

    select case(dataStyle)
    
    ! Read from a dataset
    case('set')

      ! Find the table name
      call splitSetTabName(inSetTabName, inSetName, inTabName) ! in ssclib/dal_aux
      set = dataSet(inSetName, READ)
      tab = table(set, inTabName)

      ! Allocate output array sizes
      vsize = numberOfRows(tab)
      allocate(camX(vsize), camY(vsize), ccdNum(vsize))

      ! Read from columns depending on the input units
      select case(inputUnit)

      ! Detector coordinates
      case ('det')
         call readColDataToFixed(tab, 'DETX', camX)
         call readColDataToFixed(tab, 'DETY', camY)
      case ('raw')
         call readColDataToFixed(tab, 'RAWX', camX)
         call readColDataToFixed(tab, 'RAWY', camY)
         call readColDataToFixed(tab, 'CCDNR', ccdNum)
      case default
         call error('badInpUnit', "The value of the inputunits parameter was &
                    &not recognised.")
      end select

      call release(set)

    case('user')
      allocate(camX(1), camY(1), ccdNum(1))
      camX(1) = userCamX
      camY(1) = userCamY
      ccdNum(1) = userCcd

    case default
call error('badDataStyle', "The value of the --datastyle parameter was &
&not recognised.")
    end select

  end subroutine readCamera

  !---------------------------------------------------------------------
  subroutine eSkyrawToDet(rawX, rawY, ccdNum, detX, detY)
  !---------------------------------------------------------------------

  ! Need to divide the input positions per CCD and make the transformation
  !
  ! Boring but not difficult.....
  ! For now do it one by one - should be fast anyway.
    implicit none

    real(double), intent(in) :: rawX(:),&
                                rawY(size(rawX))
    integer(int8),  intent(in) :: ccdNum(size(rawX))
    real(double), intent(out) :: detX(size(rawX)),&
                                 detY(size(rawX))
    !integer(int16) :: iRawX, iRawY
    integer(int32) :: i, ccd

    ! Check behaviour of nint
    do i = 1, size(rawX)
      
       ! Set the CCD number
       !call CAL_setState(instrument=instrumentId, ccdChipId=ccdNum(i))
       ccd = ccdNum(i)
       call CAL_setState(ccdChipId=ccd)

       ! Convert this set of coordinates from raw to detector coords
       call rawToDet(rawX(i), rawY(i), detX(i), detY(i))

    end do

  end subroutine eSkyrawToDet
  
  !---------------------------------------------------------------------
  subroutine linearToImPixels(detx, dety, wcs_image, xImagePix, yImagePix)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: detx(:),&
                                dety(size(detx))
    type(WcsT), intent(in) :: wcs_image
    real(single), intent(out) :: xImagePix(size(detx)),&
                                 yImagePix(size(detx))
    integer(int32) :: i
    real(double) :: rx, ry, lx, ly

    ! Convert each celestial pair into the pixel position in image
    do i = 1, size(detx)

    ! Use the wcs_image construct to make a linear transormation between
    ! the input and output coordinates
       rx = detx(i) - wcs_image%x%refWorld
       ry = dety(i)  - wcs_image%y%refWorld
       lx = rx / wcs_image%x%matrixOnAxis
       ly = ry / wcs_image%y%matrixOnAxis

       xImagePix(i) = lx + wcs_image%x%refPixels
       yImagePix(i) = ly + wcs_image%y%refPixels

    end do

  end subroutine linearToImPixels

  !---------------------------------------------------------------------
  subroutine getWcsKeys(calInfoSetName, wcs_image, wcs_sky)
  !---------------------------------------------------------------------
    implicit none

    character(*), intent(in) :: calInfoSetName
    type(WcsT), intent(out) :: wcs_image, wcs_sky

    type(DataSetT) :: set
    type(ArrayT) :: inArray
    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld
    integer :: status

    set = dataSet(calInfoSetName, READ)
    inArray = array(set, 0, READ)

    ! Try reading the WCS keywords from the set or the array headers
    call readWcs(set, inArray, wcs_image)

    ! Try reading the WCS REF keywords for conversion to X/Y
    call readWcsSkyKeys(calInfoSetName, wcs_sky)

  end subroutine getWcsKeys

  !---------------------------------------------------------------------
  subroutine makeOutput(rawX, rawY, offChip, ccdNum, detX, detY,&
                        ra, dec, xImagePix, yImagePix,&
                        xSkyPix, ySkyPix, instrumentId)
  !---------------------------------------------------------------------
    implicit none

    real(double), intent(in) :: rawX(:),&
                                rawY(size(rawX)),&
                                detX(size(rawX)),&
                                detY(size(rawX))
    real(double),   intent(in) :: ra(size(rawX)),&  ! In radians
                                  dec(size(rawX))
    real(single),   intent(in) :: xImagePix(size(rawX)),&  
                                  yImagePix(size(rawX)),&
                                  xSkyPix(size(rawX)),&
                                  ySkyPix(size(rawX))
    logical(bool),  intent(in) :: offChip(size(rawX))
    integer(int8),  intent(in) :: ccdNum(size(rawX))
    integer(int32), intent(in) :: instrumentId

    type(DataSetT) :: set
    type(TableT)   :: tab
    type(ColumnT)  :: rawXcol, rawYcol, ccdNrCol, flagCol
    type(ColumnT)  :: detXcol, detYcol, racol, deccol
    type(ColumnT)  :: imXcol, imYcol, skyXcol, skyYcol
    real(single), dimension(:), pointer :: dalRawX, dalRawY,&
                                   dalDetX, dalDetY, dalImX,&
                                   dalImY, dalSkyX, dalSkyY
    real(double), dimension(:), pointer :: dalRa, dalDec
    integer(int8), dimension(:), pointer :: dalCcdNr
    integer(int32), dimension(:), pointer :: dalFlag
    character(37) :: outputStr
!    integer(int16) :: i
!    character(12) :: srcNumStr
character(7) :: onChipStr

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Writing output data.')

    select case(dataStyle)

    ! Writing to output file
    case('set')
     if (withOutSet) then
        call splitSetTabName(outSetTabName, outSetName, outTabName)
        ! in ssclib/dal_aux

      if (outTabName .eq. '') call error('emptyTableName',"Parameter &
        &--outtab should be given in the form dataSetName:tableName")

      call message(AppMsg, Verbose, 'Opening Output file: '&
           //trim(outSetName)//':'//trim(outTabName) )
      set = dataSet(outSetName, CREATE)
      call setAttribute(set, 'INSTRUME', trim(instId(instrumentId)))
!*** put in DATE-TIME and wcs and *_PNT kwds too??
      tab = addTable(set, outTabName, size(rawX))

     else
      call message(AppMsg, Verbose, 'Overwriting Input file: '&
           //trim(inSetName)//':'//trim(inTabName) )
      set = dataSet(inSetName, MODIFY)
      tab = table(set, inTabName, MODIFY)
     end if

     ! Write all the RAW pixel columns
     rawXcol  = addOrOpenColumn(tab, 'RAWX',  REAL32, 'CCD pixels')
     rawYcol  = addOrOpenColumn(tab, 'RAWY',  REAL32, 'CCD pixels')
     ccdNrCol = addOrOpenColumn(tab, 'CCDNR', INTEGER8)
     flagCol  = addOrOpenColumn(tab, 'FLAG',  INTEGER32)
      ! addOrOpenColumn is in ssclib/dal_aux.

     dalRawX  => real32Data(rawXcol)
     dalRawY  => real32Data(rawYcol)
     dalCcdNr => int8Data(ccdNrCol)
     dalFlag  => int32Data(flagCol)
     dalRawX  = real(rawX,single)
     dalRawY  = real(rawY,single)
     dalCcdNr = ccdNum
     dalFlag  = 0
     where(offChip) dalFlag = ibset(dalFlag, OFF_CHIP_BIT)

     ! Write all the DET pixel columns
     detXcol = addOrOpenColumn(tab, 'DETX', REAL32)
     detYcol = addOrOpenColumn(tab, 'DETY', REAL32)
     dalDetX  => real32Data(detXcol)
     dalDetY  => real32Data(detYcol)
     dalDetX  = real(detX, single)
     dalDetY  = real(detY, single)

     ! Write all the celestial coord columns
     racol = addOrOpenColumn(tab, 'RA', REAL64)
     deccol = addOrOpenColumn(tab, 'DEC', REAL64)
     dalRa => real64Data(racol)
     dalDec  => real64Data(deccol)
     dalRa = ra / PiOn180
     dalDec = dec / PiOn180

     ! Write all the image pixel columns
     imXcol = addOrOpenColumn(tab, 'IMAGEX', REAL32)
     imYcol = addOrOpenColumn(tab, 'IMAGEY', REAL32)
     dalImX  => real32Data(imXcol)
     dalImY  => real32Data(imYcol)
     dalImX  = real(xImagePix, single)
     dalImY  = real(yImagePix, single)

     ! Write all the sky pixel columns
     skyXcol = addOrOpenColumn(tab, 'X', REAL32)
     skyYcol = addOrOpenColumn(tab, 'Y', REAL32)
     dalSkyX  => real32Data(skyXcol)
     dalSkyY  => real32Data(skyYcol)
     dalSkyX  = real(xSkyPix, single)
     dalSkyY  = real(ySkyPix, single)

      call release(set)

    case('user')
      outputStr = '' ! fills it with blanks.

      if (withHeader) then
        write(*, '(a)') '# Instrument: '//trim(instId(instrumentId))
        write(*, '(a, f10.6, a)') '# RA (deg)   DEC (deg)'
      end if

      write(*, '(f8.4, a3, f8.4, a3)') ra(1) / PiOn180, &
               '   ', dec(1) / PiOn180

! Ouput raw coordinates
      if (offChip(1)) then
        onChipStr = '      F'
call warning('notOnChip', 'The calculated raw XY position does not fall &
&on any ccd. The position relative to the nearest CCD has been printed.')
      else
        onChipStr = '      T'
      end if

      if (withHeader) then
          write(*, '(a)') '#'
          write(*, '(a)') '# rawX      rawY    ccd   On chip?'
!                         '+xxx.xx   +xxx.xx   xx      x'
      end if
!        write(*, '(i3, a8, i3, a5, i2)') rawX(1), '        '&
write(*, '(f7.2, a3, f7.2, a3, i2, a7)') rawX(1), '   '&
          , rawY(1), '   ', ccdNum(1), onChipStr

      if (withHeader) then
          write(*, '(a)') '#'
          write(*, '(a)') '# detX         detY'
!                         '+xxx.xx   +xxx.xx   xx      x'
      end if
!        write(*, '(i3, a8, i3, a5, i2)') rawX(1), '        '&
write(*, '(f10.2, a3, f10.2, a3)') detX(1), '   ', detY(1)

!*** send it as a message???

     ! Output the image and sky coords if input from a set
      if (calInfoStyle .eq. 'set') then

       if (withHeader) then
          write(*, '(a)') '#'
          write(*, '(a)') '# Image X    Y pixel'
       end if
       write(*, '(f8.2, a3, f8.2, a3)') xImagePix(1), '   ', yImagePix(1)

       if (withHeader) then
          write(*, '(a)') '#'
          write(*, '(a)') '# Sky X        Y pixel'
       end if
       write(*, '(f10.2, a3, f10.2, a3)') xSkyPix(1), '   ', ySkyPix(1)
      end if

      print *, outputStr
!*** send it as a message???

    case default
call error('badDataStyle', "The value of the --datastyle parameter &
&was not recognised.")
    end select

  end subroutine makeOutput

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module edet2sky_mod
